/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//#include <stdarg.h>
#include <limits.h>

#include "fitsio.h"
#include "my_stdio.h"
#include "local_gpib.h"
#include "camera.h"
#include "camera_setup.h"

/* Local emulated function: */
static int e_ibclr   ( int ud );
static int e_ibconfig( int ud, int option, int value );
static int e_ibdev   ( int board_index, int pad, int sad, 
		int timo, int send_eoi, int eosmode );
static int e_ibln    ( int ud, int pad, int sad, short *found_listener );
static int e_ibrd    ( int ud, void *buf, long count );
static int e_ibrsp   ( int ud, char *spr );
static int e_ibtmo   ( int ud, int v );
static int e_ibwrt   ( int ud, const void *buf, long count );
static int e_ibwrta  ( int ud, const void *buf, long count );


/* Exported functions: */
int (*l_ibclr)   ( int ud ) = e_ibclr ;
int (*l_ibconfig)( int ud, int option, int value ) = e_ibconfig;
int (*l_ibdev)   ( int board_index, int pad, int sad, 
		   int timo, int send_eoi, int eosmode ) = e_ibdev;
int (*l_ibln)    ( int ud, int pad, int sad, short *found_listener ) = e_ibln;
int (*l_ibrd)    ( int ud, void *buf, long count ) = e_ibrd;
int (*l_ibrsp)   ( int ud, char *spr ) = e_ibrsp;
int (*l_ibtmo)   ( int ud, int v ) = e_ibtmo;
int (*l_ibwrt)   ( int ud, const void *buf, long count ) = e_ibwrt;
int (*l_ibwrta)  ( int ud, const void *buf, long count ) = e_ibwrta;

volatile int ibsta = 0,
	ibcnt = 0,
	iberr = 0;


#define BUFFER_LEN 16
static char result[BUFFER_LEN];
static int read_lut_flag = 0;
static int send_data_flag = 0;

static int etime = 2;
static int treg = 1;
static int temperature = -45;
static int gain = 0;
static int x0 = 0;
static int xlen = 576;
static int y0 = 0;
static int ylen = 384;
static int auto_win = 1;
static int max_win = 254;
static int min_win = 10;
static int zoom = 0;
static char *comment = NULL;
static int mouse_x = 0;
static int mouse_y = 0;
static int mouse_v = 2048;
static int imtype = IMT_NONE;
static char base_emulate_fits[] = ".";


/* FIXME: */
static int e_ibclr( int ud )
{
	PVERB(PINFO, "ibclr:\n");
	return 0;
}


static int e_ibconfig( int ud, int option, int value )
{
	PVERB(PINFO, "ibconfig:\n");

	ibsta = 0;

	return ibsta;
}


/* FIXME: */
static int e_ibdev( int board_index, int pad, int sad, 
		int timo, int send_eoi, int eosmode )
{
	PVERB(PINFO, "ibdev:\n");

	return 1;
}


/* FIXME: */
static int e_ibln( int ud, int pad, int sad, short *found_listener )
{
	PVERB(PINFO, "ibln:\n");

	return 0;
}


static int read_scale_fits(char *image, void * buf, long count)
{
	int i, j=0;
	/*
	for(i=0; i < (count>>1);i++){
		if (j == 384) j = 0;
		*(((int16_t *)buf)+i) = j & 0xfff;
		j++;
	}
	*/
	fitsfile *fptr;   /* FITS file pointer, defined in fitsio.h */
	int status = 0;   /* CFITSIO status value MUST be initialized to zero! */
	int bitpix, naxis;
	long naxes[2] = {1,1}, fpixel[2] = {1,1};
	double *pixels;
	double factor = 1.0;
	double exposure;

	if (!fits_open_file(&fptr, image, READONLY, &status)) {

		if ( fits_read_key(fptr, TDOUBLE, "EXPTIME", &exposure,
				   NULL, &status) ) {
			PVERB(PERR, "getting: %s EXPTIME not found\n", image);
			return -1;
		}
		if ( imtype == IMT_IMAGE || imtype == IMT_DARK) {
			factor =  ((double) etime) / ((double) exposure * 10.0);
			printf("*** factor: %f\n", factor);
		}

		if (!fits_get_img_param(fptr, 2, &bitpix, &naxis, 
					naxes, &status) ) {
			if (naxis > 2 || naxis == 0) {
				PVERB(PERR,"Error: only 1D or 2D images are supported\n");
			} else {
				/* get memory for 1 row */
				pixels = (double *) malloc(naxes[0] * sizeof(double));
				
				if (pixels == NULL) {
					PVERB(PERR, "Memory allocation error\n");
					return -1;
				}
				j = count>>1;
				/* loop over all the rows in the image, top to bottom */
				for (fpixel[1] = naxes[1]; fpixel[1] >= 1; fpixel[1]--) {
					if (fits_read_pix(fptr, TDOUBLE, fpixel, naxes[0], NULL,
							  pixels, NULL, &status) )  /* read row of pixels */
						break;  /* jump out of loop on error */
					for (i = naxes[0] -1 ; i >= 0; i--) {
						j--;
						if ( imtype == IMT_IMAGE || imtype == IMT_DARK)
							*(((int16_t *)buf)+j) = ((int16_t) (pixels[i] * factor)) &0x0fff;
						else
							*(((int16_t *)buf)+j) = ((int16_t) pixels[i]) &0x0fff;
						//j++;
					}
				}
				free(pixels);
			}
		}
		fits_close_file(fptr, &status);
	}
	
	if (status) fits_report_error(stderr, status); /* print any error message */
	return(status);
	
	return 0;
}


/* FIXME: */
static int e_ibrd( int ud, void *buf, long count )
{
	int i, j=0;
	char image[_POSIX_PATH_MAX];

	PVERB(PINFO, "ibrd:\n");
	if ( send_data_flag ) {
		send_data_flag = 0;

		switch(imtype) {
		case IMT_IMAGE:
			sprintf(image, "%s/expose.fits[%d:%d,%d:%d]",
			       base_emulate_fits, y0 + 1, ylen, x0 + 1, xlen);
			read_scale_fits(image, buf, count);		
			break;
		case IMT_DARK:
			sprintf(image, "%s/dark.fits[%d:%d,%d:%d]",
			       base_emulate_fits, y0 + 1, ylen, x0 + 1, xlen);
			read_scale_fits(image, buf, count);		
			break;
		case IMT_BIAS:
			sprintf(image, "%s/bias.fits[%d:%d,%d:%d]",
			       base_emulate_fits, y0 + 1, ylen, x0 + 1, xlen);
			read_scale_fits(image, buf, count);		
			break;
		case IMT_NONE:
		default:
			for(i=0; i < (count>>1);i++){
				if (j == 384) j = 0;
				*(((char *) buf)+(i<<1)) = j & 0xff;
				*(((char *) buf)+(i<<1|1)) = (j & 0xf00) >> 8;
				j++;
			}
			break;
		}
	}else {
		memcpy(buf, result, count);
	}
	ibcnt = count;

	return 0;
}


/* FIXME: */
static int e_ibrsp( int ud, char *spr )
{
	PVERB(PINFO, "ibrsp:");
	*spr = CMD_OK;

	return 0;
}


/* FIXME: */
static int e_ibtmo( int ud, int v )
{
	PVERB(PINFO, "ibtmo:");

	return 0;
}


/* FIXME: */
static int e_ibwrt( int ud, const void *buf, long count )
{
	char type = *((char *)buf);

	PVERB(PINFO, "ibwrt: %s [%ld]\n", (char *)buf, count);

	if ( read_lut_flag ) {
		read_lut_flag = 0;
		printf("Rcv LUT: %ld bytes\n", count);
		return 0;
	}

	switch(type) {
	case '@': /* Read command */
		memset(&result, 0, BUFFER_LEN);
		switch( *((char *)buf+1) ) {
		case 'A': sprintf(result,"%d",etime); break;
		case 'B': sprintf(result,"%d",gain);  break;
		case 'C': sprintf(result,"%d",treg);  break;
		case 'D': sprintf(result,"%d",y0);    break;
		case 'E': sprintf(result,"%d",x0);    break;
		case 'F': sprintf(result,"%d",ylen);  break;
		case 'G': sprintf(result,"%d",xlen);  break;
		case 'H': sprintf(result,"%d",zoom);  break;
		case 'I': sprintf(result,"%d",auto_win); break;
		case 'J': if ( comment ) sprintf(result,"%s",comment); break;
		case 'K':
		case 'L':
		case 'M': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'N': sprintf(result,"%d",max_win);  break;
		case 'O': sprintf(result,"%d",min_win);  break;
		case 'P':
		case 'Q':
		case 'R':
		case 'S': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'T': sprintf(result,"%d",temperature); break;
		case 'U': sprintf(result,"%d",mouse_y);  break;
		case 'V': sprintf(result,"%d",mouse_x);  break;
		case 'W': sprintf(result,"%d",mouse_v);  break;
		default:
			PVERB(PERR, "Unknow R cmd: %s\n", (char *) buf);
			break;
		}
		break;
	case '!': /* Write command */
		switch( *((char *)buf+1) ) {
		case 'A': etime = atoi( ((char *)buf+2) );    break;
		case 'B': gain = atoi( ((char *)buf+2) );     break;
		case 'C': treg = atoi( ((char *)buf+2) );     break;
		case 'D': y0 = atoi( ((char *)buf+2) );       break;
		case 'E': x0 = atoi( ((char *)buf+2) );       break;
		case 'F': ylen = atoi( ((char *)buf+2) );     break;
		case 'G': xlen = atoi( ((char *)buf+2) );     break;
		case 'H': zoom = atoi( ((char *)buf+2) );     break;
		case 'I': auto_win = atoi( ((char *)buf+2) ); break;
		case 'J':
			if ( comment ) free(comment);
			comment = (char *) malloc((count - 2) * sizeof(char));
			memcpy(comment, ((char *)buf+2), count - 2);
			break;
		case 'K':
		case 'L':
		case 'M': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'N': max_win = atoi( ((char *)buf+2) );  break;
		case 'O': min_win = atoi( ((char *)buf+2) );  break;
		case 'P':
		case 'Q':
		case 'R':
		case 'S': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'U': mouse_y = atoi( ((char *)buf+2) );  break;
		case 'V': mouse_x = atoi( ((char *)buf+2) );  break;
		case 'W': mouse_v = atoi( ((char *)buf+2) );  break;
		default:
			PVERB(PERR, "Unknow W cmd: %s\n", (char *) buf);
			break;
		}
		break;
	case ':': /* Action command */
		switch( *((char *)buf+1) ) {
		case 'A': imtype = IMT_IMAGE; break;
		case 'B': imtype = IMT_DARK;  break;
		case 'C': imtype = IMT_BIAS;  break;

		case 'J':
			send_data_flag = 1;
			break;

		case 'R': /* FIXME */break;
		case 'U': 
			read_lut_flag = 1;
			break;
		default:
			PVERB(PERR, "Unknow Action cmd: %s\n", (char *) buf);
			break;
		}
		break;
	default:
		//iberr = 1;
		PVERB(PERR, "Unknow cmd: %s\n", (char *) buf);
	}

	return 0;
}


static int e_ibwrta( int ud, const void *buf, long count )
{
	PVERB(PINFO, "ibwrta:");

	return e_ibwrt( ud, buf, count );
}

